/*
 * Copyright (c) 2009 Maxim Bourmistrov <maxim@unixconn.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/systm.h>
#include <mach/mach_types.h>
#include <mach/kmod.h>

#include <sys/kpi_mbuf.h>
#include <netinet/kpi_ipfilter.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <sys/errno.h>

#define NKE_NAME	"trafscrambler"
#define NKE_VERSION	"0.1"
#define MBUF_TAG	0xd33d

#warning NUM_FAKE_SYN defined in Makefile

static ipfilter_t filter_ref;

static mbuf_tag_id_t gidtag;
static mbuf_t mbuf_out = NULL;

static int syn_seen = 0;

typedef enum PACKETPROCFLAGS {
	INBOUND_DONE = 1,
	OUTBOUND_DONE
} PACKETPROCFLAGS;

enum {
	FAKE_CKSUM,
	FAKE_SEQ,
	SYN_DECOY,
	ZERO_WIN
};

#if DUMP_PKT
static void
dump_mbuf(mbuf_t m)
{
	char *frame;
	int i;
	
	frame = mbuf_data(m);
	// dumps only proto 6 (TCP)
	do {
		printf("=======mbuf dump=====\n");
		for (i = 0; i < mbuf_len(m); i++) {	
			printf("%d:[%02X] ", i, (u_int8_t)frame[i]);
		}
		printf("\n======end of dump======\n");
		m = mbuf_next(m);

	} while (m != NULL && frame[9] == 6);

	return;
}
#endif

static inline long
getrandom (int min, int max)
{
	return ((random () % (int) (((max) + 1) - (min))) + (min));
}

static int 
check_tag(mbuf_t m, mbuf_tag_id_t id_tag,
		mbuf_tag_type_t tag_type, PACKETPROCFLAGS value)
{
	errno_t status;
	int *tag_ref;
	size_t len;
	
	/* Check whether we have seen this packet before. */
	status = mbuf_tag_find(m, id_tag,
				tag_type, &len, (void**)&tag_ref);
	if ((status == 0) && (*tag_ref == value)
				&& (len == sizeof(value)))
		return 1;
		
	return 0;
}

static errno_t
set_tag(mbuf_t m, mbuf_tag_id_t id_tag,
		mbuf_tag_type_t tag_type, PACKETPROCFLAGS value)
{	
	errno_t status;
	int *tag_ref = NULL;
	size_t len;
	
	/* look for existing tag */
	status = mbuf_tag_find(m, id_tag,
				tag_type, &len, (void*)&tag_ref);
	if (status != 0) {		
		status = mbuf_tag_allocate(m, id_tag, tag_type,
						sizeof(value),
						MBUF_WAITOK,
						(void**)&tag_ref);
		if (status == 0)
			*tag_ref = value;
	}

	return status;
}

static void
inject_mbuf(mbuf_t m)
{
	(void)ipf_inject_output(m, filter_ref, 0);
}

static void
inject(int th_flag, int attck_type)
{
	mbuf_t mbuf_copy = NULL;
	char *frame;

	mbuf_dup(mbuf_out, MBUF_WAITOK, &mbuf_copy);

	if (mbuf_copy == NULL || mbuf_len(mbuf_copy) <= 0)
		return;

	frame = mbuf_data(mbuf_copy);

#if INFO
	printf("%s: Injected ", NKE_NAME);
#endif
	switch(attck_type) {
	case FAKE_CKSUM: {
			/* th_sum */
			frame[36] = frame[36] + (u_int8_t)random();
			frame[37] = frame[37] + (u_int8_t)random();

			if (th_flag == TH_SYN) {
				frame[33] = (th_flag|TH_URG);
#if INFO
				printf("SYN");
#endif
			}
			else if (th_flag == TH_FIN) {
				frame[33] = (th_flag|TH_ACK);
				frame[28] = frame[29] = 0x00;
				frame[30] = frame[31] = 0x00;
#if INFO
				printf("FIN");
#endif
			}
			else if (th_flag == TH_RST) {
				frame[33] = (th_flag|TH_ACK);
				frame[28] = frame[29] = 0x00;
				frame[30] = frame[31] = 0x00;
#if INFO
				printf("RST");
#endif
			}
			else if (th_flag == TH_ACK) {
				frame[33] = th_flag;
				/* th_ack seq */
				frame[30] = (u_int8_t)random();
				frame[31] = (u_int8_t)random();
#if INFO
				printf("ACK");
#endif
			}
#if INFO
			else 
				printf("original(th_flag: 0x%02x)", (u_int8_t)frame[33]);

			printf(" pkt with wrong CKSUM\n");
#endif
		}
		break;
	case FAKE_SEQ: {
			/* th_seq */
			frame[24] = frame[25] = (u_int8_t)random();
			frame[26] = frame[27] = (u_int8_t)random();

			/* th_ack seq */
			if (th_flag != TH_SYN && th_flag != 0xdead) {
				frame[28] = frame[29] = (u_int8_t)random();
				frame[30] = frame[31] = (u_int8_t)random();
			}

			if (th_flag == TH_SYN) {
				frame[33] = (th_flag|TH_URG);
#if INFO
				printf("SYN");
#endif
			}
			else if (th_flag == TH_FIN) {
				frame[33] = (th_flag|TH_ACK);
#if INFO
				printf("FIN");
#endif
			}
			else if (th_flag == TH_RST) {
				frame[33] = (th_flag|TH_ACK);
#if INFO
				printf("RST");
#endif
			}
			else if (th_flag == TH_ACK) {
				frame[33] = th_flag;
#if INFO
				printf("ACK");
#endif
			}
#if INFO
			else
				printf("original(th_flag: 0x%02x)", (u_int8_t)frame[33]);

			printf(" pkt with wrong SEQ\n");
#endif
		}
		break;
	case ZERO_WIN:
			{
				frame[24] = frame[25] = (u_int8_t)random();
				frame[26] = frame[27] = (u_int8_t)random();

				frame[28] = frame[29] = (u_int8_t)random();
				frame[30] = frame[31] = (u_int8_t)random();

				frame[33] = (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG);
				frame[34] = frame[35] = 0x00; /* th_win */
#if INFO
				printf("pkt with ZERO window\n");
#endif
			}
			break;
	case SYN_DECOY: {
			/* TTL */
			frame[8] = getrandom(10, 30);

			/* src IP */
			frame[12] = getrandom(100, 200);
			frame[13] = getrandom(100, 200);
			frame[14] = getrandom(100, 200);
			frame[15] = getrandom(100, 200);

			if ((random() % 5) == 0) {
				frame[33] |= TH_ACK;
				frame[30] = (u_int8_t)random();
				frame[31] = (u_int8_t)random();
			}
				
#if INFO
			printf("%d spoofed fake SYN(s)\n", NUM_FAKE_SYN);
#endif
		}
		break;
	default:
		return;
	}

	/* random ip_id */
	frame[4] = frame[4] + (u_int8_t)(random() % 10);
	frame[5] = frame[5] + (u_int8_t)(random() % 10);

	if (set_tag(mbuf_copy, gidtag, MBUF_TAG, OUTBOUND_DONE) != 0) {
#if DEBUG
		printf("%s(): failed to set TAG. freeing.\n", __FUNCTION__);
#endif
		mbuf_freem(mbuf_copy);
		return;
	}

	inject_mbuf(mbuf_copy);
	return;
}

static errno_t
ipv4_output_func_cb(void* cookie, mbuf_t *data, ipf_pktopts_t options)
{
	mbuf_t m = *data;
	char *frame = NULL;
#if USE_SYN_DECOY
	int i;
#endif

	/* no use of mbuf we can't modify */
	if (mbuf_type(m) != MBUF_TYPE_HEADER)
		return 0;

	/* no use of mbuf we can't tag */
	if ((mbuf_flags(m) & MBUF_PKTHDR) != MBUF_PKTHDR)
		return 0;

	/* Don't process already seen mbuf */
	if (check_tag(m, gidtag, MBUF_TAG, OUTBOUND_DONE)) {
#if DEBUG
		printf("%s(): mbuf already seen\n", __FUNCTION__);
#endif
		return 0;
	}

	frame = mbuf_data(m);	

	/* TCP
	 *
	 */
	if (frame != NULL && (u_int8_t)frame[9] == 6) {

		switch((u_int8_t)(frame[33] & ~(TH_URG|TH_ECE|TH_CWR))) {
			case TH_SYN:
			{
#if DEBUG
				printf("Got SYN\n");
#endif
				/* Stash away original mbuf */
				if ((mbuf_dup(m, MBUF_WAITOK, &mbuf_out)) != 0)
					return 0;

				/* Inject number of fake SYN pkts */
#if USE_SYN_DECOY
				for (i = 0; i < NUM_FAKE_SYN; i++)
					(void)inject(0xdead, SYN_DECOY);
#endif
				(void)inject(0xdead, ZERO_WIN);
				(void)inject(TH_ACK, FAKE_SEQ);
				(void)inject(TH_RST, FAKE_SEQ);
				/* Pre-connection SYN */
				(void)inject(0xdead, FAKE_CKSUM);
				
				syn_seen = 1;
			}
			break;
			case TH_ACK:
			{
#if DEBUG
				printf("Got ACK\n");
#endif
				if ((random() % 5) == 0) {
					if ((mbuf_dup(m, MBUF_WAITOK, &mbuf_out)) != 0)
						return 0;

					(void)inject(TH_SYN, FAKE_SEQ);
					(void)inject(TH_RST, FAKE_SEQ);
					(void)inject(0xdead, ZERO_WIN);
				}
			}
			break;
			case (TH_PUSH|TH_ACK):
			{
#if DEBUG
				printf("Got PUSH/ACK\n");
#endif
				if (syn_seen) {
					syn_seen = 0; /* switch off */

					/* TCP reset attack
					 * http://kerneltrap.org/node/3072
					 */
					(void)inject(TH_RST, FAKE_SEQ);

					(void)inject(TH_ACK, FAKE_SEQ);
					/* Post-connection SYN */
					(void)inject(TH_SYN, FAKE_SEQ);

				}
			}
			break;
			case (TH_FIN|TH_ACK):
			{
#if DEBUG
				printf("Got FIN/ACK\n");
#endif
				if ((mbuf_dup(m, MBUF_WAITOK, &mbuf_out)) != 0)
					return 0;

				(void)inject(TH_RST, FAKE_SEQ);
				(void)inject(0xdead, ZERO_WIN);
#if USE_SYN_DECOY
				(void)inject(0xdead, SYN_DECOY);
#endif
			}
			break;
			case (TH_SYN|TH_ACK):
			case TH_FIN:
			case TH_RST:
			default: break;
		}

#if DUMP_PKT
	dump_mbuf(m);
#endif
	}

	return 0;
}

/* ----------------------------------------------------------------------------------------------------------------*/

static struct ipf_filter ipv4_filter = {
	NULL,
	NKE_NAME,
	NULL,
	ipv4_output_func_cb,
	NULL
};

/*
 * start function. called when the kext gets loaded.
 */
static kern_return_t
trafscrambler_module_start(struct kmod_info *ki, void *data)
{
	int ret;

	ret = mbuf_tag_id_find(NKE_NAME, &gidtag);
	if (ret != 0) {
		printf("mbuf_tag_id_find returned error %d\n", ret);
		goto out;
	}
	
	ret = ipf_addv4(&ipv4_filter, &filter_ref);
	if (ret != 0)
		goto out;

	printf("%s(%s) was successfully loaded!\n", NKE_NAME, NKE_VERSION);
	return KERN_SUCCESS;

out:
	printf("%s(%s) failed to load!\n", NKE_NAME, NKE_VERSION);
	return KERN_FAILURE;
}

/*
 * stop function. called when the kext should be unloaded. unloading can be prevented by
 * returning failure
 */
static kern_return_t
trafscrambler_module_stop(struct kmod_info *ki, void *data)
{
	errno_t ret;

	ret = ipf_remove(filter_ref);
	if (ret != 0)
		printf("%s(%s): Failed uninstall ipfiler\n", NKE_NAME, NKE_VERSION);

	if (mbuf_out != NULL)
		(void)mbuf_freem(mbuf_out);

	printf("%s(%s) was succsessfully unloaded!\n", NKE_NAME, NKE_VERSION);
	return KERN_SUCCESS;
}

KMOD_DECL(trafscrambler, NKE_VERSION);

